#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Script:
    circle_nums.py
Description:
    Manim captcha scene of numbers in a matrix distribution animation.
Author:
    Jose Miguel Rios Rubio
Date:
    27/02/2026
Version:
    1.1.1
"""

###############################################################################
# Libraries
###############################################################################

# Standard Libraries
import random
import secrets

# Third-Party Libraries
import manim


###############################################################################
# Scene
###############################################################################

class MatrixNums(manim.Scene):
    '''
    Manim captcha scene of numbers in a matrix distribution animation.
    '''

    theme_default = {
        "bg_color": manim.BLACK,
        "draw_color": manim.WHITE,
        "selector_color": manim.BLUE_D,
        "container_color": manim.BLACK,
    }

    theme_dark = {
        "bg_color": manim.ManimColor("#0E1621"),
        "draw_color": manim.WHITE,
        "selector_color": manim.BLUE_D,
        "container_color": manim.ManimColor("#1F1F1F")
    }

    theme_light = {
        "bg_color": manim.ManimColor("#6FA788"),
        "draw_color": manim.BLACK,
        "selector_color": manim.RED_D,
        "container_color": manim.WHITE,
    }

    def __init__(self,
                 captcha_code: str | int | None = None,
                 properties: dict | None = None,
                 **kwargs):
        super().__init__(**kwargs)
        self.captcha_code = captcha_code
        self.noise = False
        self._apply_theme(self.theme_default)
        # Configure properties
        if properties:
            if "noise" in properties:
                self.noise = properties["noise"]
            if "theme" in properties:
                if properties["theme"] == "dark":
                    self._apply_theme(self.theme_dark)
                elif properties["theme"] == "light":
                    self._apply_theme(self.theme_light)
                elif properties["theme"] == "random":
                    list_themes = list(
                        [self.theme_default, self.theme_dark, self.theme_light]
                    )
                    random_theme = secrets.choice(list_themes)
                    self._apply_theme(random_theme)
                else:
                    self._apply_theme(self.theme_default)
            if "bg_color" in properties:
                self.bg_color = properties["bg_color"]
            if "draw_color" in properties:
                self.draw_color = properties["draw_color"]
            if "selector_color" in properties:
                self.selector_color = properties["selector_color"]
            if "container_color" in properties:
                self.container_color = properties["container_color"]
        # Apply background color
        self.camera.background_color = self.bg_color

    def construct(self):
        SELECTOR_OPACITY = 0.5
        SELECTOR_SIZE = 1.2
        NUMBERS_SIZE = 84
        # Ensure the code is an unsigned number (int or str)
        if not self._is_valid_captcha_code(self.captcha_code):
            return
        self.captcha_code = str(self.captcha_code)
        # Random 0-9 Numbers
        num_map = {}
        orden = secrets.SystemRandom().sample(range(10), 10)
        nums = manim.VGroup()
        for n in orden:
            t = manim.Text(str(n), font_size=NUMBERS_SIZE,
                           color=self.draw_color)
            nums.add(t)
            num_map[n] = t
        # Numbers distribution
        ROWS = 2
        COLS = 5
        SPACING_X = 2.5
        SPACING_Y = 2.5
        for i, num in enumerate(nums):
            row = i // COLS
            col = i % COLS
            x = (col - (COLS-1)/2) * SPACING_X
            y = ((ROWS-1)/2 - row) * SPACING_Y
            num.move_to(manim.np.array([x, y, 0]))
        # Draw Container
        container = manim.Rectangle(width=12,
                                    height=5,
                                    stroke_width=10,
                                    color=self.draw_color,
                                    fill_color=self.container_color,
                                    fill_opacity=1.0)
        self.add(container)
        # Display noise
        if self.noise:
            self._add_noise(num_lines=25, num_dots=35, width=13, height=6)
        # Draw Selector
        selector = manim.Square(side_length=SELECTOR_SIZE,
                                color=self.draw_color,
                                fill_color=self.selector_color,
                                fill_opacity=SELECTOR_OPACITY)
        self.add(selector)
        # Display numbers with fade-in
        for num in nums:
            self.play(manim.FadeIn(num, scale=0.1), run_time=0.1)
        # Display Selector transition over numbers
        l_target_numbers = [int(d) for d in self.captcha_code]
        for target_num in l_target_numbers:
            # Go to target number
            target = num_map[target_num]
            self.play(
                selector.animate.move_to(target.get_center()),
                run_time=0.6,
                rate_func=manim.smooth
            )
            self.wait(0.2)
            # Move Selector to original position
            self.play(
                selector.animate.move_to(manim.ORIGIN),
                run_time=0.6,
                rate_func=manim.smooth
            )
        self.wait(1)

    def _add_noise(self, num_lines=20, num_dots=30, width=16, height=6):
        '''Add noise particles free moving around the image.'''
        DOTS_RADIUS_MIN_MAX = (0.04, 0.08)
        LINES_STROKE_WIDTH_MIN_MAX = (3, 8)
        noise_group = manim.VGroup()
        # Slim Lines
        for _ in range(num_lines):
            start = manim.np.array([
                random.uniform(-width/2, width/2),
                random.uniform(-height/2, height/2),
                0
            ])
            end = start + manim.np.array([
                random.uniform(-0.5, 0.5),
                random.uniform(-0.5, 0.5),
                0
            ])
            stroke_width = random.uniform(
                LINES_STROKE_WIDTH_MIN_MAX[0], LINES_STROKE_WIDTH_MIN_MAX[1])
            line = manim.Line(
                start, end, stroke_width=stroke_width, color=self.draw_color)
            line._velocity = manim.np.array([
                random.uniform(-0.02, 0.02),
                random.uniform(-0.02, 0.02),
                0
            ])
            noise_group.add(line)
        # Small Dots
        for _ in range(num_dots):
            pos = manim.np.array([
                random.uniform(-width/2, width/2),
                random.uniform(-height/2, height/2),
                0
            ])
            radius = random.uniform(
                DOTS_RADIUS_MIN_MAX[0], DOTS_RADIUS_MIN_MAX[1])
            dot = manim.Dot(point=pos, radius=radius, color=self.draw_color)
            dot._velocity = manim.np.array([
                random.uniform(-0.03, 0.03),
                random.uniform(-0.03, 0.03),
                0
            ])
            noise_group.add(dot)
        self.add(noise_group)
        # Noise movement animation updater
        def move_noise_smooth(mob, dt):
            for obj in mob:
                # Move and reflect at borders
                obj.shift(obj._velocity)
                x, y, _ = obj.get_center()
                if x < -width/2 or x > width/2:
                    obj._velocity[0] *= -1
                if y < -height/2 or y > height/2:
                    obj._velocity[1] *= -1
        noise_group.add_updater(move_noise_smooth)

    def _is_valid_captcha_code(self, captcha_code: str | int | None):
        valid = False
        if captcha_code:
            try:
                int_captcha_code = int(captcha_code)
                if int_captcha_code >= 0:
                    valid = True
            except ValueError:
                valid = False
        return valid

    def _apply_theme(self, theme):
        self.bg_color = theme["bg_color"]
        self.draw_color = theme["draw_color"]
        self.selector_color = theme["selector_color"]
        self.container_color = theme["container_color"]

###############################################################################



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Script:
    manim_captcha_generator.py
Description:
    Manim video Captcha generator.
    This component allows you to generate a captcha video file using
    Manim library according to the specified parameters.
    The captcha scene to be used in the captcha can be provided as
    custom scene or define from some builtin predefined scenes.
Author:
    Jose Miguel Rios Rubio
Date:
    14/02/2026
Version:
    1.2.0
"""

###############################################################################
# Standard Libraries
###############################################################################

import inspect
import json
import logging
import secrets
import shutil
import subprocess
import sys

from pathlib import Path
from traceback import format_exc


###############################################################################
# Third-Party Libraries
###############################################################################

import manim

from manim.constants import RendererType

###############################################################################
# Local Libraries
###############################################################################

from .colors import CaptchaColor
from .data import CaptchaData
from .scenes import CaptchaScene

###############################################################################
# Logger Setup
###############################################################################

logger = logging.getLogger(__name__)


###############################################################################
# Class ManimCaptchaGenerator
###############################################################################

class CaptchaGenerator:
    """Manim captcha Generator."""

    ###########################################################################

    ### Constants ###

    MANIM_LAUNCHER = Path(__file__).parent / "manim_launcher.py"


    ###########################################################################

    ### Data Types ###

    # None

    ###########################################################################

    ### Constructor ###

    def __init__(self):
        self.available: bool = self._is_manim_available()
        if self.available:
            logging.getLogger("manim").setLevel(logging.ERROR)

    ###########################################################################

    ### Public Methods ###

    def is_available(self):
        '''Get Generator availability.'''
        return self.available

    def generate(self,
                 code: str | None = None,
                 scene: type[manim.Scene] | None = None,
                 out_dir: Path | None = None,
                 tmp_dir: Path | None = None,
                 width: int = 854,
                 height: int = 480,
                 fps: int = 30,
                 format: str = "mp4",
                 renderer: RendererType = RendererType.CAIRO,
                 properties: dict | None = None) -> CaptchaData:
        """
        Generate a captcha for the provided code.
        Arguments:
        - scene: Manim captcha scene to use.
        - code: Captcha code to generate (string).
        - out_dir: Output directory to put the generated captcha file.
        - tmp_dir: Temporary directory for captcha build.
        - width: Output captcha resolution width.
        - height: Output captcha resolution height.
        - fps: Output captcha video framerate.
        - properties: Captcha properties (i.e. colors).
        - format: Captcha output format (mp4/gif).
        - rendered: Video renderer to use (cairo/opengl).
        """
        captcha_result = CaptchaData()
        # Do nothing if Manim is not available
        if not self.available:
            captcha_result.error = True
            captcha_result.error_info = "Manim not found in the system"
            return captcha_result
        # Use a random captcha code number if was not provided
        if code is None:
            code = self._generate_random_code()
        # Use a random captcha scene if was not provided
        if scene is None:
            scene = CaptchaScene.get_random_scene()
        # Set output captcha file (use current path if was not provided)
        if out_dir is None:
            out_dir = Path(".")
        out_file_name = f"{code}.{format}"
        out_file = out_dir / f"{code}.{format}"
        # Use current path for temporary directory if was not provided
        if tmp_dir is None:
            tmp_dir = out_dir / "tmp"
        # Set background color
        bg_color = CaptchaColor.BLACK
        if properties:
            bg_color = properties.get("bg_color", CaptchaColor.BLACK)
        # Create output directory if it doesn't exists
        out_dir.mkdir(parents=True, exist_ok=True)
        # Run Manim to generate the captcha
        result = self._run_manim(code, scene, tmp_dir, out_file_name, width,
                                  height, fps, format, renderer, bg_color,
                                  properties)
        if result.returncode != 0:
            captcha_result.error = True
            captcha_result.error_info = f"Manim process fail: {result.stderr}"
        else:
            generated_file = tmp_dir / f"videos/{height}p{fps}/{out_file_name}"
            if generated_file.exists():
                self._move_file(generated_file, out_file)
                captcha_result.code = code
                captcha_result.file = out_file
                logger.info("Generated captcha: %s", out_file)
            else:
                captcha_result.error = True
                captcha_result.error_info = "Captcha creation fail"
        self._rmdir(tmp_dir)
        return captcha_result

    ###########################################################################

    ### Private Methods ###

    def _run_manim(self, code: str, scene, media_dir: Path, out_file: Path,
                   width: int, height: int, fps: int, format: str,
                   renderer: RendererType, bg_color: str,
                   properties: dict | None):
        manim_data = {
            "scene_file": str(self._get_scene_file(scene)),
            "scene_class": scene.__name__,
            "code": code,
            "properties": properties or {},
            "renderer": renderer.name,
            "format": format,
            "pixel_width": width,
            "pixel_height": height,
            "frame_rate": fps,
            "background_color": bg_color,
            "media_dir": str(media_dir),
            "output_file": str(out_file),
        }
        json_manim_data = json.dumps(manim_data)
        # Run Manim subprocess
        cmd = [sys.executable, str(self.MANIM_LAUNCHER), json_manim_data]
        result = subprocess.run(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True
        )
        return result

    def _generate_random_code(self, digits: int = 4) -> str:
        return "".join(secrets.choice("0123456789") for _ in range(digits))

    def _rmdir(self, dir):
        try:
            shutil.rmtree(dir, ignore_errors=True)
        except Exception:
            logger.error(format_exc())
            logger.error("Fail to remove directory: %s", dir)

    def _move_file(self, file, target_dir):
        try:
            shutil.move(file, target_dir)
        except Exception:
            logger.error(format_exc())
            logger.error("Fail to move file to directory: %s -> %s",
                         file, target_dir)

    def _get_scene_file(self, scene_class):
        file_path = inspect.getfile(scene_class)
        return Path(file_path).resolve()

    def _is_manim_available(self) -> bool:
        manim_bin = shutil.which("manim")
        if not manim_bin:
            return False
        return True

###############################################################################



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Script:
    test_captcha_creation.py
Description:
    Manim Captcha creation tests.
Author:
    Jose Miguel Rios Rubio
Date:
    27/02/2026
Version:
    1.0.1
"""

###############################################################################
# Libraries
###############################################################################

# Standard Libraries
import shutil
from pathlib import Path
from traceback import format_exc

# Manim Library
from manim.constants import RendererType
from manim_captcha.generator import CaptchaGenerator
from manim_captcha.scenes import CaptchaScene
# from custom_captcha.test import TheScene as TestScene


###############################################################################
# Constants
###############################################################################

BASE_DIR = Path(__file__).resolve().parent
BUILD_DIR = BASE_DIR / "../build/captchas"


###############################################################################
# Auxiliary Functions
###############################################################################

def rmdir(dir):
    try:
        shutil.rmtree(dir, ignore_errors=True)
    except Exception:
        print(format_exc())
        print("Fail to remove directory: %s", dir)


def show_result(captcha):
    if captcha.error:
        print(captcha.error_info)
        print("Fail to create the captcha")
        return
    print("")
    print("Captcha successfully created")
    print(f"  Code: \"{captcha.code}\"")
    print(f"  File: {captcha.file}")
    print("")


###############################################################################
# Test Functions
###############################################################################

def test_captcha_creation_invalid_code():
    OUT_DIR = BUILD_DIR / "invalid"
    rmdir(OUT_DIR)
    generator = CaptchaGenerator()
    # Invalid Generation 1: Negative integer for circle nums captcha
    captcha = generator.generate(
        -123, scene=CaptchaScene.CIRCLE_NUMS,  # type: ignore[arg-type]
        out_dir=OUT_DIR)
    show_result(captcha)
    assert captcha.error
    # Invalid Generation 2: Negative number for circle nums captcha
    captcha = generator.generate("-123", scene=CaptchaScene.CIRCLE_NUMS,
                                 out_dir=OUT_DIR)
    show_result(captcha)
    assert captcha.error
    # Invalid Generation 3: Not number for circle nums captcha
    captcha = generator.generate("1AB2", scene=CaptchaScene.CIRCLE_NUMS,
                                 out_dir=OUT_DIR)
    show_result(captcha)
    assert captcha.error


def test_captcha_creation_default():
    OUT_DIR = BUILD_DIR / "default"
    NUM_CAPTCHAS_TO_GENERATE = 3
    rmdir(OUT_DIR)
    generator = CaptchaGenerator()
    # Default Generation
    for _ in range(NUM_CAPTCHAS_TO_GENERATE):
        captcha = generator.generate(out_dir=OUT_DIR)
        show_result(captcha)
        assert not captcha.error


def test_captcha_creation_short_long():
    OUT_DIR = BUILD_DIR / "short_long"
    rmdir(OUT_DIR)
    generator = CaptchaGenerator()
    # Cairo Renderer Generation
    captcha = generator.generate("5", out_dir=OUT_DIR)
    show_result(captcha)
    assert not captcha.error
    # OpenGL Renderer Generation
    captcha = generator.generate("0123456789", out_dir=OUT_DIR)
    show_result(captcha)
    assert not captcha.error


def test_captcha_creation_render():
    OUT_DIR = BUILD_DIR / "render"
    rmdir(OUT_DIR)
    generator = CaptchaGenerator()
    # Cairo Renderer Generation
    captcha = generator.generate("1234", scene=CaptchaScene.CIRCLE_NUMS,
                                 renderer=RendererType.CAIRO, out_dir=OUT_DIR)
    show_result(captcha)
    assert not captcha.error
    # OpenGL Renderer Generation
    captcha = generator.generate("5678", scene=CaptchaScene.CIRCLE_NUMS,
                                 renderer=RendererType.OPENGL, out_dir=OUT_DIR)
    show_result(captcha)
    assert not captcha.error


def test_captcha_creation_builtin_scenes():
    OUT_DIR = BUILD_DIR / "builtin_scenes"
    rmdir(OUT_DIR)
    generator = CaptchaGenerator()
    # Captcha Generation - Circle Nums
    captcha = generator.generate("0123", scene=CaptchaScene.CIRCLE_NUMS,
                                 out_dir=OUT_DIR)
    show_result(captcha)
    assert not captcha.error
    # Captcha Generation - Matrix Nums
    captcha = generator.generate("5678", scene=CaptchaScene.MATRIX_NUMS,
                                 out_dir=OUT_DIR)
    show_result(captcha)
    assert not captcha.error
    # Captcha Generation - Piramid Nums
    captcha = generator.generate("9012", scene=CaptchaScene.PIRAMID_NUMS,
                                 out_dir=OUT_DIR)
    show_result(captcha)
    assert not captcha.error


def test_captcha_creation_builtin_themes():
    OUT_DIR = BUILD_DIR / "themes"
    rmdir(OUT_DIR)
    generator = CaptchaGenerator()
    # Captcha Generation - Default Theme
    properties = {"theme": "default"}
    captcha = generator.generate("1234", scene=CaptchaScene.CIRCLE_NUMS,
                                 properties=properties, out_dir=OUT_DIR)
    show_result(captcha)
    assert not captcha.error
    # Captcha Generation - Dark Theme
    properties = {"theme": "dark"}
    captcha = generator.generate("5678", scene=CaptchaScene.CIRCLE_NUMS,
                                 properties=properties, out_dir=OUT_DIR)
    show_result(captcha)
    assert not captcha.error
    # Captcha Generation - Light Theme
    properties = {"theme": "light"}
    captcha = generator.generate("9012", scene=CaptchaScene.CIRCLE_NUMS,
                                 properties=properties, out_dir=OUT_DIR)
    show_result(captcha)
    assert not captcha.error


def test_captcha_creation_custom_theme():
    THEME = {
        "bg_color": "#84D5E9",
        "draw_color": "#FDA552",
        "selector_color": "#FDC189",
        "container_color": "#FBFCC0"
    }
    OUT_DIR = BUILD_DIR / "theme_custom"
    rmdir(OUT_DIR)
    generator = CaptchaGenerator()
    properties = THEME
    # Captcha Generation - Circle Nums
    captcha = generator.generate("1234", scene=CaptchaScene.CIRCLE_NUMS,
                                 properties=properties, out_dir=OUT_DIR)
    show_result(captcha)
    assert not captcha.error
    # Captcha Generation - Matrix Nums
    captcha = generator.generate("5678", scene=CaptchaScene.MATRIX_NUMS,
                                 properties=properties, out_dir=OUT_DIR)
    show_result(captcha)
    assert not captcha.error
    # Captcha Generation - Piramid Nums
    captcha = generator.generate("9012", scene=CaptchaScene.PIRAMID_NUMS,
                                 properties=properties, out_dir=OUT_DIR)
    show_result(captcha)
    assert not captcha.error


def test_captcha_creation_with_noise():
    OUT_DIR = BUILD_DIR / "noise"
    rmdir(OUT_DIR)
    generator = CaptchaGenerator()
    properties = {
        "theme": "random",
        "noise": True
    }
    # Captcha Generation - Circle Nums
    captcha = generator.generate("1234", scene=CaptchaScene.CIRCLE_NUMS,
                                 properties=properties, out_dir=OUT_DIR)
    show_result(captcha)
    assert not captcha.error
    # Captcha Generation - Matrix Nums
    captcha = generator.generate("2345", scene=CaptchaScene.MATRIX_NUMS,
                                 properties=properties, out_dir=OUT_DIR)
    show_result(captcha)
    assert not captcha.error
    # Captcha Generation - Piramid Nums
    captcha = generator.generate("3456", scene=CaptchaScene.PIRAMID_NUMS,
                                 properties=properties, out_dir=OUT_DIR)
    show_result(captcha)
    assert not captcha.error

###############################################################################



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Script:
    scene.py
Description:
    Manim builtin captcha scenes.
Author:
    Jose Miguel Rios Rubio
Date:
    14/02/2026
Version:
    1.0.0
"""

###############################################################################
# Libraries
###############################################################################

# Standard Libraries
import secrets

# Third-Party Libraries
import manim

# Captcha Builtin Scenes
from .scene.circle_nums import CircleNums
from .scene.matrix_nums import MatrixNums
from .scene.piramid_nums import PiramidNums
# ...


###############################################################################
# Captcha Builtin Scenes
###############################################################################

class CaptchaScene:
    '''Captcha builtin scenes.'''

    CIRCLE_NUMS = CircleNums
    MATRIX_NUMS = MatrixNums
    PIRAMID_NUMS = PiramidNums
    # ...

    @staticmethod
    def get_random_scene() -> type[manim.Scene]:
        scenes = [
            v for v in vars(CaptchaScene).values()
            if isinstance(v, type) and issubclass(v, manim.Scene)
        ]
        return secrets.choice(scenes)

###############################################################################



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Script:
    auto_generator.py
Description:
    Automatic captcha generation manager.
    This component act as a producer that periodically creates and store
    random Manim Captchas into a directory of the filesystem, allowing
    external consumers (users of the library) to request any of these
    captchas. It keeps control of the number of generated captchas, and
    apply some kind of "files rotation" of the storage, removing and
    replacing older captchas with newer ones to control and limit the
    maximum number of captchas in the filesytem.
Author:
    Jose Miguel Rios Rubio
Date:
    27/02/2026
Version:
    1.2.0
"""

###############################################################################
# Standard Libraries
###############################################################################

import asyncio
import logging
import secrets

from pathlib import Path
from traceback import format_exc


###############################################################################
# Third-Party Libraries
###############################################################################

from manim import Scene
from manim.constants import RendererType


###############################################################################
# Local Libraries
###############################################################################

from .data import CaptchaData
from .generator import CaptchaGenerator


###############################################################################
# Logger Setup
###############################################################################

logger = logging.getLogger(__name__)


###############################################################################
# Class ManimCaptchaGenerator
###############################################################################

class CaptchaAutoGenerator:
    """Automatic captcha generation manager."""

    ###########################################################################

    ### Constants ###

    # None

    ###########################################################################

    ### Data Types ###

    # None

    ###########################################################################

    ### Constructor ###

    def __init__(self,
                 out_dir: Path = Path("./captchas"),
                 interval_s: int = 60,
                 max_items: int = 100):
        '''
        Automatic captcha generation manager construction.
        Arguments:
        - out_dir: Output directory to place the generated captchas.
        - interval: Time for creation of each new Captcha (in seconds).
        - max_items: Maximum number of Captchas allowed in the
        filesystem (once this number of captchas are generated, the file
        rotate mechanism will be replacing older captchas with newer
        ones).
        '''
        self.list_captcha_files: list = list()
        self.list_scenes: list = list()
        self.out_dir = out_dir
        self.interval_s = interval_s
        self.max_items = max_items
        self.format: str = "mp4"
        self._task_create: asyncio.Task | None = None
        self._task_remove: asyncio.Task | None = None
        self._running: bool = False
        self._lock = asyncio.Lock()
        self._generator = CaptchaGenerator()
        self.available: bool = self._generator.is_available()

    ###########################################################################

    ### Public Methods ###

    def add_captcha_scene(self,
                          scene: type[Scene],
                          properties: dict | None = None,
                          width: int = 854,
                          height: int = 480,
                          fps: int = 30,
                          format: str = "mp4",
                          renderer: RendererType = RendererType.CAIRO):
        '''Add a scene to the list of captchas to be generated.'''
        captcha_cfg = {
            "scene": scene,
            "width": width,
            "height": height,
            "fps": fps,
            "format": format,
            "renderer": renderer,
            "properties": properties
        }
        if captcha_cfg not in self.list_scenes:
            self.list_scenes.append(captcha_cfg)

    async def start(self) -> bool:
        '''Launch the manager to start creating captchas.'''
        # Check if the generator can be started
        if not self.available:
            return False
        if self._running:
            return False
        # Get list of captcha files already in the output directory
        list_files = list(self.out_dir.glob(f"*.{self.format}"))
        self.list_captcha_files = sorted(
            list_files, key=lambda f: f.stat().st_mtime)
        # Start Producer and Remover tasks
        self._running = True
        self._task_create = asyncio.create_task(self._producer_loop())
        self._task_remove = asyncio.create_task(self._remove_loop())
        return True

    async def stop(self):
        '''Stop the manager and the captcha creation.'''
        if self._task_create:
            self._task_create.cancel()
            try:
                await self._task_create
            except asyncio.CancelledError:
                pass
            self._task_create = None
        if self._task_remove:
            self._task_remove.cancel()
            try:
                await self._task_remove
            except asyncio.CancelledError:
                pass
            self._task_remove = None
        self._running = False

    def is_running(self) -> bool:
        '''Get Captcha Generator manager running status.'''
        return self._running

    def num_captchas(self) -> int:
        '''
        Get the number of generated captcha files availables in the
        file system.
        '''
        return len(self.list_captcha_files)

    def get_captcha(self) -> CaptchaData:
        '''Return a random captcha file from storage.'''
        captcha_result = CaptchaData()
        if not self.available:
            captcha_result.error = True
            captcha_result.error_info = "Manim not found in the system"
            return captcha_result
        try:
            if not self.list_captcha_files:
                captcha_result.error = True
                captcha_result.error_info = "None captcha files generated"
                return captcha_result
            captcha_result.file = secrets.choice(self.list_captcha_files)
            captcha_result.code = captcha_result.file.stem
        except Exception:
            logger.error(format_exc())
            captcha_result.error = True
            captcha_result.error_info = "Fail to get captcha from filesystem"
        return captcha_result

    ###########################################################################

    ### Private Methods - Create Captchas ###

    async def _producer_loop(self):
        """Internal async captcha creation producer loop."""
        while self._running:
            try:
                async with self._lock:
                    await self._generate_captcha()
            except asyncio.CancelledError:
                raise
            except Exception:
                logger.error(format_exc())
            await asyncio.sleep(self.interval_s)

    async def _generate_captcha(self):
        """Generate a captcha."""
        loop = asyncio.get_running_loop()
        # Generate a random one from the list of configured captchas
        if len(self.list_scenes) > 0:
            captcha_cfg = secrets.choice(self.list_scenes)
            result = await loop.run_in_executor(
                None,
                self._generator.generate,
                None,  # code
                captcha_cfg["scene"],
                self.out_dir,
                None,  # tmp_dir
                captcha_cfg["width"],
                captcha_cfg["height"],
                captcha_cfg["fps"],
                captcha_cfg["format"],
                captcha_cfg["renderer"],
                captcha_cfg["properties"],
            )
        # If none captchas configured, generate one from builtin
        else:
            result = await loop.run_in_executor(
                None,
                self._generator.generate,
                None,  # code
                None,  # scene
                self.out_dir
            )
        if result.error:
            logger.error("Captcha generation fail: %s", result.error_info)
        else:
            logger.debug("Generated captcha: %s", result.file)
            self.list_captcha_files.append(result.file)

    ###########################################################################

    ### Private Methods - Remove Captchas ###

    async def _remove_loop(self):
        """Internal async captcha remove/rotate loop."""
        while self._running:
            try:
                async with self._lock:
                    self._process_remove()
            except asyncio.CancelledError:
                raise
            except Exception:
                logger.error(format_exc())
            await asyncio.sleep(self.interval_s)

    def _process_remove(self):
        """Remove oldest file if limit exceeded."""
        # Dont remove if number of captchas is less than max_items limit
        if len(self.list_captcha_files) < self.max_items:
            return
        # Remove oldest captcha file
        while len(self.list_captcha_files) > self.max_items:
            oldest = self.list_captcha_files.pop(0)
            try:
                # Atomic rename (protect race condition from consumers)
                # and delete the file
                deleting = oldest.with_suffix(oldest.suffix + ".deleting")
                oldest.rename(deleting)
                deleting.unlink(missing_ok=True)
                logger.info("Removed old captcha: %s", oldest)
            except PermissionError:
                logger.warning("File in use, skipping removal: %s", oldest)
            except Exception:
                logger.error(format_exc())

###############################################################################



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Script:
    manim_launcher.py
Description:
    Manim process launcher.
    This script runs Manim process to render a Scene for the provided
    JSON configuration. It main purpose is to be used by the generator
    via subprocess for memory isolation (running Manim as subprocess
    ensures all RAM memory resources are completely free once process
    ends) to avoid RAM memory fragmentation or any possible leak at
    Manim or ffmpg affects the generator process that uses it.
Author:
    Jose Miguel Rios Rubio
Date:
    27/02/2026
Version:
    1.0.0
"""

###############################################################################
# Standard Libraries
###############################################################################

import importlib.util
import json
import logging
import sys


###############################################################################
# Third-Party Libraries
###############################################################################

import manim


###############################################################################
# Main Function
###############################################################################

def main(argc, argv):
    # Set Manim logging level to error to hide messages
    logging.getLogger("manim").setLevel(logging.ERROR)
    # Get Manim config data from argments
    if argc < 1:
        return 1
    json_data = json.loads(argv[0])
    # Sanitize json data to expected types
    if len(json_data["properties"]) == 0:
        json_data["properties"] = None
    bg_color = manim.ManimColor(json_data["background_color"])
    json_data["background_color"] = bg_color
    json_data["renderer"] = manim.constants.RendererType[json_data["renderer"]]
    print(json_data)
    # Load Manim Scene
    scene_file = json_data["scene_file"]
    class_name = json_data["scene_class"]
    spec = importlib.util.spec_from_file_location("custom_scene", scene_file)
    module = importlib.util.module_from_spec(spec)
    sys.modules["custom_scene"] = module
    spec.loader.exec_module(module)
    SceneClass = getattr(module, class_name)
    # Setup config renderer type
    manim_config = {
        "renderer": json_data["renderer"],
        "format": json_data["format"],
        "pixel_width": json_data["pixel_width"],
        "pixel_height": json_data["pixel_height"],
        "frame_rate": json_data["frame_rate"],
        "background_color": json_data["background_color"],
        "media_dir": json_data["media_dir"],
        "output_file": json_data["output_file"],
        # "disable_caching": True,
        "write_to_movie": True,
        "progress_bar": "none",
        "preview": False,
    }
    with manim._config.tempconfig(manim_config):
        SceneClass(json_data["code"], json_data["properties"]).render()
    return 0


###############################################################################
# Main Call
###############################################################################

if __name__ == "__main__":
    rc = main(len(sys.argv) - 1, sys.argv[1:])
    sys.exit(rc)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Script:
    test_captcha_auto_generator.py
Description:
    Manim Captcha Automatic Generator tests.
Author:
    Jose Miguel Rios Rubio
Date:
    27/02/2026
Version:
    1.0.1
"""

###############################################################################
# Libraries
###############################################################################

# Standard Libraries
import asyncio
import shutil
from pathlib import Path
from traceback import format_exc

# Manim Library
from manim_captcha.auto_generator import CaptchaAutoGenerator


###############################################################################
# Constants
###############################################################################

BASE_DIR = Path(__file__).resolve().parent
BUILD_DIR = BASE_DIR / "../build/captchas/auto_generation"


###############################################################################
# Auxiliary Functions
###############################################################################

def rmdir(dir):
    try:
        shutil.rmtree(dir, ignore_errors=True)
    except Exception:
        print(format_exc())
        print("Fail to remove directory: %s", dir)


def show_result(captcha):
    if captcha.error:
        print("Fail to get captcha:")
        print(captcha.error_info)
        return
    print("")
    print("Captcha successfully obtained")
    print(f"  Code: \"{captcha.code}\"")
    print(f"  File: {captcha.file}")
    print("")


###############################################################################
# Test Functions
###############################################################################

async def test_captcha_auto_generator():
    MAX_NUM_CAPTCHAS = 3
    TIME_GEN_INTERVAL_S = 10
    TIME_CAPTCHA_GEN_S = 12
    auto_generator = CaptchaAutoGenerator(
        BUILD_DIR, TIME_GEN_INTERVAL_S, MAX_NUM_CAPTCHAS)
    # Test get captcha returns error when no captchas are availables
    rmdir(BUILD_DIR)
    captcha = auto_generator.get_captcha()
    show_result(captcha)
    assert captcha.error
    # Test Start of captcha automatic generator doesn't fail
    result = await auto_generator.start()
    assert result
    # Test number of generated captchas is 0 right after start
    num_captchas = auto_generator.num_captchas()
    assert num_captchas == 0
    # Test getting a captcha after wait for first generation interval
    await asyncio.sleep(TIME_CAPTCHA_GEN_S)
    num_captchas = auto_generator.num_captchas()
    assert num_captchas == 1
    captcha = auto_generator.get_captcha()
    show_result(captcha)
    assert not captcha.error
    # Test getting a captcha after wait for second generation interval
    await asyncio.sleep(TIME_CAPTCHA_GEN_S)
    num_captchas = auto_generator.num_captchas()
    assert num_captchas == 2
    captcha = auto_generator.get_captcha()
    show_result(captcha)
    assert not captcha.error
    # Test getting a captcha after wait for third generation interval
    await asyncio.sleep(TIME_CAPTCHA_GEN_S)
    num_captchas = auto_generator.num_captchas()
    assert num_captchas == 3
    captcha = auto_generator.get_captcha()
    show_result(captcha)
    # Test number of captchas after some asyncio is the configured limit
    await asyncio.sleep(3 * TIME_CAPTCHA_GEN_S)
    num_captchas = auto_generator.num_captchas()
    assert num_captchas == MAX_NUM_CAPTCHAS
    captcha = auto_generator.get_captcha()
    show_result(captcha)
    # Test Generator stop
    running = auto_generator.is_running()
    assert running
    await auto_generator.stop()
    running = auto_generator.is_running()
    assert not running

###############################################################################
